
div {
    clear: both;
    margin: 0 0 0 1em;
    padding: 0.2em 0;
    font: 10pt sans;
}


.Change .description {
    margin: 0;
    font: 12pt sans;
}


.is_unchanged {
    color: green;
}


.is_changed {
    color: red;
}


.is_ignored {
    color: orange;
}


.vertical {
    clear: none;
    float: left;
}


.vertical.ldata, .vertical.rdata {
    width: auto;
    min-width: 20em;
    max-width: 47%;
    padding: 0 2em 1em 0;
    margin: 0;
}


.lrdata {
    min-width: 20em;
    padding: 0 2em 1em 0;
    margin: 0 2em;
}


table {
    border-collapse: collapse;
    color: black;
}


td {
    vertical-align: top;
    border: thin solid grey;
    padding: 0.2em 1em;
}


td.const-value {
    /* max-width: 24em; */
    word-wrap: break-word;
    font-family: "Verdana", "monospace";
}


th {
    color: black;
    vertical-align: top;
    padding: 0 1em;
}


table.left-headers th {
    text-align: right;
}


ul.breadcrumbs li {
    border: thin solid grey;
    display: inline;
    list-style-type: none;
    padding: 0 0.5em;
    margin: 0;
}


ul.breadcrumbs a {
    text-decoration: none;
}


/* The end. */
